#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "types.h"
#include "cpu.h"

#include "allegro.h"

#include "regs.h"
#include "modrm.h"
#include "access.h"
#include "data8.h"
#include "data16.h"
#include "arith8.h"
#include "arith16.h"
#include "logic8.h"
#include "logic16.h"
#include "shift8.h"
#include "shift16.h"
#include "mult8.h"
#include "mult16.h"
#include "branch8.h"
#include "branch16.h"
#include "string8.h"
#include "string16.h"
#include "flag.h"
#include "disasm.h"

// CPU core variables
byte op1_8,op2_8;
byte aux8;
word op1_16,op2_16;
word aux16;
dword op1_32,op2_32;
dword aux32;

// Debug variables
word IPnext;

inline byte GetOF() { return (EFLAGS&0x00000800)>>11; }
inline byte GetDF() { return (EFLAGS&0x00000400)>>10; }
inline byte GetIF() { return (EFLAGS&0x00000200)>>9;  }
inline byte GetSF() { return (EFLAGS&0x00000080)>>7;  }
inline byte GetZF() { return (EFLAGS&0x00000040)>>6;  }
inline byte GetAF() { return (EFLAGS&0x00000010)>>4;  }
inline byte GetCF() { return (EFLAGS&0x00000004)>>2;  }
inline byte GetPF() { return EFLAGS&0x00000001;       }

#define SetOF EFLAGS|=0x00000800
#define SetDF EFLAGS|=0x00000400
#define SetIF EFLAGS|=0x00000200
#define SetSF EFLAGS|=0x00000080
#define SetZF EFLAGS|=0x00000040
#define SetAF EFLAGS|=0x00000010
#define SetCF EFLAGS|=0x00000004
#define SetPF EFLAGS|=0x00000001

#define ClrOF EFLAGS&=0xFFFFF7FF
#define ClrDF EFLAGS&=0xFFFFFBFF
#define ClrIF EFLAGS&=0xFFFFFDFF
#define ClrSF EFLAGS&=0xFFFFFF7F
#define ClrZF EFLAGS&=0xFFFFFFBF
#define ClrAF EFLAGS&=0xFFFFFFEF
#define ClrCF EFLAGS&=0xFFFFFFFB
#define ClrPF EFLAGS&=0xFFFFFFFE

#define Is_CF(R,OP)      (R<OP)
#define Is_AF(R,OP1,OP2) (((OP1^OP2^R)&0x10)==0x10)
#define Is_SF8(R)        (R>=0x80)
#define Is_SF16(R)       (R>=0x8000)
#define Is_SF32(R)       (R>=0x80000000)
#define Is_ZF(R)         (!R)
#define Is_PF(R)         (PARITY[R&0xFF])

byte PARITY[256];  // Tableau pour le flag PF

void ResetCPU(unsigned long ProgramSize)
{
 EAX=EDX=EBX=0x00000000;
 EBP=ESI=EDI=0x00000000;
 ECX=ProgramSize;
 ESP=0x0000FFFE;
 CS=SS=DS=ES=FS=GS=0x0000;
 EIP=0x00000100;
 EFLAGS=0x00000002;
}

void Lire_Params(char *params) 
{
 char *p;
 int nbcar;

 nbcar=1; // Un espace est forcement inclus
 p=params;
 while(*p) {
  if (*p != 0x0A) {
   nbcar++;
   PCMemory[(CS<<4)+0x080+nbcar]=*p;
  }
  p++;
 }
 PCMemory[(CS<<4)+0x080]=nbcar;
}

inline byte CPUReadMemory8()
{
 return PCMemory[(CS<<4)+(IP++)];
}

inline word CPUReadMemory16()
{
 return CPUReadMemory8()+(CPUReadMemory8()<<8);
}

inline dword CPUReadMemory32()
{
 return CPUReadMemory16()+(CPUReadMemory16()<<16);
}

void CPUInitParity()
{
 int i,j,count;

 for(i=0; i<256; i++)
 {
  count=0;      
  for (j=i; j>0; j>>=1)
   if (j&1) count++;
  PARITY[i]=!(count&1);
 }
}

void DisasmData8(char *dest,byte data)
{
 char NewData[256];
 sprintf(NewData,"%02X",data);
 strcat(dest,NewData);
}

void DisasmData16(char *dest,word data)
{
 char NewData[256];
 sprintf(NewData,"%02X%02X",data&0xFF,data>>8);
 strcat(dest,NewData);
}

void DisasmData32(char *dest,dword data)
{
 char NewData[256];
 sprintf(NewData,"%02X%02X%02X%02X",data&0xFF,(data>>8)&0xFF,(data>>16)&0xFF,data>>24);
 strcat(dest,NewData);
}

void DebugCPUDecodeModRM(byte modrm)
{
 Mod=modrm>>6;
 Regop=(modrm>>3)&0x07;
 Rm=modrm&0x07;
}

void StrReplace(char *dest,char *src,char *pos,int len,char *replace)
{
 char *psrc=src;
 char *preplace=replace;
 char *pdest=dest;
 while (*psrc) {
  if (psrc==pos) {
   while (*preplace) {
    *pdest++=*preplace++;
   }
   psrc+=len;
  } else
   *pdest++=*psrc++;
 }
 *pdest++='\0';
}

void DebugCPUDisasm()
{
 char Opcode[256];
 char OpcodeR[256];
 char Replace[256];
 char Datas[256];
 byte data8;
 word data16;
 dword data32;
 char *locate;
 int pass;
 int Finpass;

 Datas[0]='\0';
 data8=CPUReadMemory8();
 DisasmData8(Datas,data8);
 strcpy(Opcode,DisasmOpcodes[data8]);
 strcpy(OpcodeR,Opcode);

 if ((locate=strstr(Opcode,"gr")) || (locate=strstr(Opcode,"rm")))
 {
  // Must decode Mod R/M byte
  data8=CPUReadMemory8();
  DisasmData8(Datas,data8);
  DebugCPUDecodeModRM(data8);
 }

 if (locate=strstr(Opcode,"gr1")) {
  strcpy(Replace,DisasmGrp1[Regop]);
  StrReplace(OpcodeR,Opcode,locate,3,Replace);
  strcpy(Opcode,OpcodeR); 
 } else if (locate=strstr(Opcode,"gr2")) {
  strcpy(Replace,DisasmGrp2[Regop]);
  StrReplace(OpcodeR,Opcode,locate,3,Replace);
  strcpy(Opcode,OpcodeR); 
 } else if (locate=strstr(Opcode,"gr3")) {
  strcpy(Replace,DisasmGrp3[Regop]);
  StrReplace(OpcodeR,Opcode,locate,3,Replace);
  strcpy(Opcode,OpcodeR); 
 } else if (locate=strstr(Opcode,"gr4")) {
  strcpy(Replace,DisasmGrp4[Regop]);
  StrReplace(OpcodeR,Opcode,locate,3,Replace);
  strcpy(Opcode,OpcodeR); 
 } else if (locate=strstr(Opcode,"gr5")) {
  strcpy(Replace,DisasmGrp5[Regop]);
  StrReplace(OpcodeR,Opcode,locate,3,Replace);
  strcpy(Opcode,OpcodeR); 
 }

 if (locate=strstr(Opcode,"rm8")) {
  switch(Mod) {
   case 0x00 : strcpy(Replace,DisasmMod0[Rm]); break;
   case 0x01 : strcpy(Replace,DisasmMod1[Rm]); break;
   case 0x02 : strcpy(Replace,DisasmMod2[Rm]); break;
   case 0x03 : strcpy(Replace,DisasmReg8[Rm]); break;
  }
  StrReplace(OpcodeR,Opcode,locate,3,Replace);
  strcpy(Opcode,OpcodeR); 
 } else if (locate=strstr(Opcode,"rm16")) {
  switch(Mod) {
   case 0x00 : strcpy(Replace,DisasmMod0[Rm]); break;
   case 0x01 : strcpy(Replace,DisasmMod1[Rm]); break;
   case 0x02 : strcpy(Replace,DisasmMod2[Rm]); break;
   case 0x03 : strcpy(Replace,DisasmReg16[Rm]); break;
  }
  StrReplace(OpcodeR,Opcode,locate,4,Replace);
  strcpy(Opcode,OpcodeR); 
 }

 if (locate=strstr(Opcode,"r8")) {
  strcpy(Replace,DisasmReg8[Regop]);
  StrReplace(OpcodeR,Opcode,locate,2,Replace);
  strcpy(Opcode,OpcodeR); 
 } else if (locate=strstr(Opcode,"r16")) {
  strcpy(Replace,DisasmReg16[Regop]);
  StrReplace(OpcodeR,Opcode,locate,3,Replace);
  strcpy(Opcode,OpcodeR); 
 }

 // Two passes : imm8/imm16 in Mod R/M adresses and imm8/imm16 in opcode
 Finpass=0; pass=0;
 while ((pass<2)&&(!Finpass)) {
  if (locate=strstr(Opcode,"imm8")) {
   data8=CPUReadMemory8();
   DisasmData8(Datas,data8);
   sprintf(Replace,"%02X",data8);
   StrReplace(OpcodeR,Opcode,locate,4,Replace);
   strcpy(Opcode,OpcodeR);
  } else if (locate=strstr(Opcode,"imm16")) {
   data16=CPUReadMemory16();
   DisasmData16(Datas,data16);
   sprintf(Replace,"%04X",data16);
   StrReplace(OpcodeR,Opcode,locate,5,Replace);
   strcpy(Opcode,OpcodeR);
  } else
   Finpass=1;
  pass++;
 }

 if (locate=strstr(Opcode,"rel8")) {
  data8=CPUReadMemory8();
  DisasmData8(Datas,data8);
  sprintf(Replace,"%04X",IP+(signed char)data8);
  StrReplace(OpcodeR,Opcode,locate,4,Replace);
  strcpy(Opcode,OpcodeR);
 } else if (locate=strstr(Opcode,"rel16")) {
  data16=CPUReadMemory16();
  DisasmData16(Datas,data16);
  sprintf(Replace,"%04X",IP+(signed int)data16);
  StrReplace(OpcodeR,Opcode,locate,5,Replace);
  strcpy(Opcode,OpcodeR);
 }

 if (locate=strstr(Opcode,"seg")) {
  strcpy(Replace,DisasmSegReg[Regop]);
  StrReplace(OpcodeR,Opcode,locate,3,Replace);
  strcpy(Opcode,OpcodeR); 
 }

 printf("%12-s  %s\n",Datas,OpcodeR);
}

void DebugCPULine()
{
 printf("%04X:%04X ",CS,IP);
}

void NextIP()
{
 IP=IPnext;
}

void DebugCPUShowRegisters()
{
 printf("\n");
 printf("EAX=%08X    EBX=%08X    ECX=%08X    EDX=%08X\n",EAX,EBX,ECX,EDX);
 printf("ESP=%08X    EBP=%08X    ESI=%08X    EDI=%08X\n",ESP,EBP,ESI,EDI);
 printf("CS=%04X SS=%04X DS=%04X ES=%04X FS=%04X GS=%04X EIP=%08X\n",CS,SS,DS,ES,FS,GS,EIP);
 printf("OF DF IF SF ZF AF PF CF\n");
 printf(" %d  %d  %d  %d  %d  %d  %d  %d\n",GetOF(),GetDF(),GetIF(),GetSF(),GetZF(),GetAF(),GetPF(),GetCF());
 printf("\n");
}

void DumpMEM()
{
 int i,j;
 word OldIP=IP;
 char Datas[256];
 char Ascii[256];
 byte data8;
 char car[2];

 for(i=0;i<8;i++) {
  Datas[0]='\0';
  Ascii[0]='\0';
  DebugCPULine();
  for(j=0;j<16;j++) {
   data8=CPUReadMemory8();
   DisasmData8(Datas,data8); strcat(Datas," ");
   car[0]=(data8<32)?' ':(char)data8;
   car[1]='\0';
   strcat(Ascii,car);
  }
  printf("%s  %s\n",Datas,Ascii);
 }
 IP=OldIP;
}


void DebugCPU()
{
 int i;
 word OldIP=IP;

 for(i=0;i<8;i++) {
  if (!i) {
   printf(">");
   DebugCPULine();
   DebugCPUDisasm();
   IPnext=IP;
  } else {
   printf(" ");
   DebugCPULine();
   DebugCPUDisasm();
  }
 }
 IP=OldIP;
 DebugCPUShowRegisters();
}

void gr1_rm8_imm8()
{
 DecodeModRM;
 switch(Regop) {
  case 0x00 : ADD_rm8_imm8; break;
  case 0x01 : OR_rm8_imm8; break;
  case 0x04 : AND_rm8_imm8; break;
  case 0x05 : SUB_rm8_imm8; break;
  case 0x06 : XOR_rm8_imm8; break;
  case 0x07 : CMP_rm8_imm8; break;
  default   : printf("ERROR : gr1_rm8_imm8 Regop=%02X\n",Regop); exit(0); break;
 }
}

void gr1_rm16_imm16()
{
 DecodeModRM;
 switch(Regop) {
  case 0x00 : ADD_rm16_imm16; break;
  case 0x01 : OR_rm16_imm16; break;
  case 0x04 : AND_rm16_imm16; break;
  case 0x05 : SUB_rm16_imm16; break;
  case 0x06 : XOR_rm16_imm16; break;
  case 0x07 : CMP_rm16_imm16; break;
  default   : printf("ERROR : gr1_rm16_imm16 Regop=%02X\n",Regop); exit(0); break;
 }
}

void gr1_rm16_imm8()
{
 DecodeModRM;
 switch(Regop) {
  case 0x00 : ADD_rm16_imm8; break;
  case 0x01 : OR_rm16_imm8; break;
  case 0x04 : AND_rm16_imm8; break;
  case 0x05 : SUB_rm16_imm8; break;
  case 0x06 : XOR_rm16_imm8; break;
  case 0x07 : CMP_rm16_imm8; break;
  default   : printf("ERROR : gr1_rm16_imm8 Regop=%02X\n",Regop); exit(0); break;
 }
}

void gr2_rm8(int count)
{
 DecodeModRM;
 switch(Regop) {
  case 0x00 : ROL_rm8(count); break;
  case 0x01 : ROR_rm8(count); break;
  case 0x02 : RCL_rm8(count); break;
  case 0x03 : RCR_rm8(count); break;
  case 0x04 : SHL_rm8(count); break;
  case 0x05 : SHR_rm8(count); break;
  case 0x07 : SAR_rm8(count); break;
  default   : printf("ERROR : gr2_rm8(%d) Regop=%02X\n",count,Regop); exit(0); break;
 }
}

void gr2_rm16(int count)
{
 DecodeModRM;
 switch(Regop) {
  case 0x00 : ROL_rm16(count); break;
  case 0x01 : ROR_rm16(count); break;
  case 0x02 : RCL_rm16(count); break;
  case 0x03 : RCR_rm16(count); break;
  case 0x04 : SHL_rm16(count); break;
  case 0x05 : SHR_rm16(count); break;
  case 0x07 : SAR_rm16(count); break;
  default   : printf("ERROR : gr2_rm16(%d) Regop=%02X\n",count,Regop); exit(0); break;
 }
}

void gr3_rm8()
{
 DecodeModRM;
 switch(Regop) {
//  case 0x00 : TEST_rm8; break;
  case 0x02 : NOT_rm8; break;
  case 0x03 : NEG_rm8; break;
  case 0x04 : MUL_rm8; break;
  case 0x05 : IMUL_rm8; break;
  case 0x06 : DIV_rm8; break;
  case 0x07 : IDIV_rm8; break;
  default   : printf("ERROR : gr3_rm8 Regop=%02X\n",Regop); exit(0); break;
 }
}

void gr3_rm16()
{
 DecodeModRM;
 switch(Regop) {
//  case 0x00 : TEST_rm16; break;
  case 0x02 : NOT_rm16; break;
  case 0x03 : NEG_rm16; break;
  case 0x04 : MUL_rm16; break;
  case 0x05 : IMUL_rm16; break;
  case 0x06 : DIV_rm16; break;
  case 0x07 : IDIV_rm16; break;
  default   : printf("ERROR : gr3_rm16 Regop=%02X\n",Regop); exit(0); break;
 }
}

void gr4()
{
 DecodeModRM;
 switch(Regop) {
  case 0x00 : INC_rm8; break;
  case 0x01 : DEC_rm8; break;
  default   : printf("ERROR : gr4 Regop=%02X\n",Regop); exit(0); break;
 }
}

void gr5()
{
 DecodeModRM;
 switch(Regop) {
  case 0x00 : INC_rm16; break;
  case 0x01 : DEC_rm16; break;
  default   : printf("ERROR : gr5 Regop=%02X\n",Regop); exit(0); break;
 }
}

void Interrupt_DOS()
{
 dword offset;
 char data;
 int finchaine;

 switch(AH) {
  case 0x09 : offset=(DS<<4)+DX;
	      finchaine=0;
	      while(!finchaine) {
	       data=PCMemory[offset++];
	       if (data!=0x24)
		printf("%c",data);
	       else
		finchaine=1;
	      }
	      break;
  case 0x4C : exit(1); // Fin normale du programme
	      break;
  default   : printf("ERROR : MS-DOS interrupt function %02X not supported\n",AH); exit(0); break;
 }
}


void Interrupt_VIDEO()
{
 switch(AH) {
  case 0x00 : switch(AL) {
	       case 0x03 : set_gfx_mode (GFX_TEXT, 80, 25, 0, 0); break;
	       case 0x13 : set_gfx_mode (GFX_AUTODETECT, 320, 200, 0, 0); break;
	      }
	      break;
  default   : printf("ERROR : VIDEO interrupt function %02X not supported\n",AH); exit(0); break;
 }
}

void Interrupt()
{
 byte data8=CPUReadMemory8();
 switch(data8) {
  case 0x10 : Interrupt_VIDEO(); break;
  case 0x20 : exit(0); break;
  case 0x21 : Interrupt_DOS(); break;
  default   : printf("ERROR : interrupt %02X not supported\n",data8); exit(0); break;
 }
}

void EmuleCPU_Bis()
{
 byte data8=CPUReadMemory8();
 switch(data8) {
  default   :
	  printf("ERROR : unknown opcode 0F%02X\n",data8);
	  exit(0);
	  break;
 }
}

void EmuleCPU()
{
 byte data8=CPUReadMemory8();
 switch(data8) {
  case 0x00 : ADD_rm8_r8;       break;  
  case 0x01 : ADD_rm16_r16;     break;  
  case 0x02 : ADD_r8_rm8;       break;
  case 0x03 : ADD_r16_rm16;     break;
  case 0x04 : ADD_AL_imm8;      break;
  case 0x05 : ADD_AX_imm16;     break;
  case 0x06 : PUSH(ES);         break;
  case 0x07 : POP(ES);          break;
  case 0x08 : OR_rm8_r8;        break;  
  case 0x09 : OR_rm16_r16;      break;  
  case 0x0A : OR_r8_rm8;        break;
  case 0x0B : OR_r16_rm16;      break;
  case 0x0C : OR_AL_imm8;       break;
  case 0x0D : OR_AX_imm16;      break;
  case 0x0E : PUSH(CS);         break;
  case 0x0F : EmuleCPU_Bis();   break;
  case 0x16 : PUSH(SS);         break;
  case 0x17 : POP(SS);          break;
  case 0x1E : PUSH(DS);         break;
  case 0x1F : POP(DS);          break;
  case 0x20 : AND_rm8_r8;       break;  
  case 0x21 : AND_rm16_r16;     break;  
  case 0x22 : AND_r8_rm8;       break;
  case 0x23 : AND_r16_rm16;     break;
  case 0x24 : AND_AL_imm8;      break;
  case 0x25 : AND_AX_imm16;     break;
  case 0x28 : SUB_rm8_r8;       break;  
  case 0x29 : SUB_rm16_r16;     break;  
  case 0x2A : SUB_r8_rm8;       break;
  case 0x2B : SUB_r16_rm16;     break;
  case 0x2C : SUB_AL_imm8;      break;
  case 0x2D : SUB_AX_imm16;     break;
  case 0x30 : XOR_rm8_r8;       break;  
  case 0x31 : XOR_rm16_r16;     break;  
  case 0x32 : XOR_r8_rm8;       break;
  case 0x33 : XOR_r16_rm16;     break;
  case 0x34 : XOR_AL_imm8;      break;
  case 0x35 : XOR_AX_imm16;     break;
  case 0x40 : INC_r16(AX);      break;
  case 0x41 : INC_r16(CX);      break;
  case 0x42 : INC_r16(DX);      break;
  case 0x43 : INC_r16(BX);      break;
  case 0x44 : INC_r16(SP);      break;
  case 0x45 : INC_r16(BP);      break;
  case 0x46 : INC_r16(SI);      break;
  case 0x47 : INC_r16(DI);      break;
  case 0x48 : DEC_r16(AX);      break;
  case 0x49 : DEC_r16(CX);      break;
  case 0x4A : DEC_r16(DX);      break;
  case 0x4B : DEC_r16(BX);      break;
  case 0x4C : DEC_r16(SP);      break;
  case 0x4D : DEC_r16(BP);      break;
  case 0x4E : DEC_r16(SI);      break;
  case 0x4F : DEC_r16(DI);      break;
  case 0x50 : PUSH(AX);         break;
  case 0x51 : PUSH(CX);         break;
  case 0x52 : PUSH(DX);         break;
  case 0x53 : PUSH(BX);         break;
  case 0x54 : PUSH(SP);         break;
  case 0x55 : PUSH(BP);         break;
  case 0x56 : PUSH(SI);         break;
  case 0x57 : PUSH(DI);         break;
  case 0x58 : POP(AX);          break;
  case 0x59 : POP(CX);          break;
  case 0x5A : POP(DX);          break;
  case 0x5B : POP(BX);          break;
  case 0x5C : POP(SP);          break;
  case 0x5D : POP(BP);          break;
  case 0x5E : POP(SI);          break;
  case 0x5F : POP(DI);          break;
  case 0x70 : JO_rel8;          break;
  case 0x71 : JNO_rel8;         break;
  case 0x72 : JB_rel8;          break;
  case 0x73 : JNB_rel8;         break;
  case 0x74 : JZ_rel8;          break;
  case 0x75 : JNZ_rel8;         break;
  case 0x76 : JBE_rel8;         break;
  case 0x77 : JNBE_rel8;        break;
  case 0x78 : JS_rel8;          break;
  case 0x79 : JNS_rel8;         break;
  case 0x7A : JP_rel8;          break;
  case 0x7B : JNP_rel8;         break;
  case 0x7C : JL_rel8;          break;
  case 0x7D : JNL_rel8;         break;
  case 0x7E : JLE_rel8;         break;
  case 0x7F : JNLE_rel8;        break;
  case 0x80 : gr1_rm8_imm8();   break;
  case 0x81 : gr1_rm16_imm16(); break;
  case 0x83 : gr1_rm16_imm8();  break;
  case 0x88 : MOV_rm8_r8;       break;
  case 0x89 : MOV_rm16_r16;     break;
  case 0x8A : MOV_r8_rm8;       break;
  case 0x8B : MOV_r16_rm16;     break;
  case 0x8C : MOV_rm16_seg;     break;
  case 0x8E : MOV_seg_rm16;     break;
  case 0xAC : LODSB;            break;
  case 0xB0 : MOV_Rb_imm8(AL);  break;
  case 0xB1 : MOV_Rb_imm8(CL);  break;
  case 0xB2 : MOV_Rb_imm8(DL);  break;
  case 0xB3 : MOV_Rb_imm8(BL);  break;
  case 0xB4 : MOV_Rb_imm8(AH);  break;
  case 0xB5 : MOV_Rb_imm8(CH);  break;
  case 0xB6 : MOV_Rb_imm8(DH);  break;
  case 0xB7 : MOV_Rb_imm8(BH);  break;
  case 0xB8 : MOV_Rw_imm16(AX); break;
  case 0xB9 : MOV_Rw_imm16(CX); break;
  case 0xBA : MOV_Rw_imm16(DX); break;
  case 0xBB : MOV_Rw_imm16(BX); break;
  case 0xBC : MOV_Rw_imm16(SP); break;
  case 0xBD : MOV_Rw_imm16(BP); break;
  case 0xBE : MOV_Rw_imm16(SI); break;
  case 0xBF : MOV_Rw_imm16(DI); break;
  case 0xCD : Interrupt();      break;
  case 0xD0 : gr2_rm8(1);       break;
  case 0xD1 : gr2_rm16(1);      break;
  case 0xD2 : gr2_rm8(CL);      break;
  case 0xD3 : gr2_rm16(CL);     break;
  case 0xE9 : JMP_rel16;        break;
  case 0xF5 : CMC;              break;
  case 0xF6 : gr3_rm8();        break;
  case 0xF7 : gr3_rm16();       break;
  case 0xF8 : CLC;              break;
  case 0xF9 : STC;              break;
  case 0xFA : CLI;              break;
  case 0xFB : STI;              break;
  case 0xFC : CLD;              break;
  case 0xFD : STD;              break;
  case 0xFE : gr4();            break;
  default   :
	  printf("ERROR : unknown opcode %02X\n",data8);
	  exit(0);
	  break;
 }
}



